/**********************************************************************/
//
// FILE:    adscam.h
//          ADS camera class.
// CREATED: 15 Feb 04  (ds)
// CHANGES:
// REFS:    ADS40 Info Kit
//
// (c) 2004 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#ifndef DEP_ADSCAM_INCLUDED
#define DEP_ADSCAM_INCLUDED

#include "pbcamcal.h"

using namespace std;
using namespace GenLib2;


namespace DEP
{

/** ADS camera class.
 *
 *  @author Dirk Stallmann
 */
class ADSCamera : public PBCameraCalib
{
public:
	/// keyword enumerator
	enum ekey
	{
		/// unknown
		UNKNOWN,

		/// version number of the calibration file
		ADS_CALIBRATION_FILE,

		/// name of calibration source
		CALIBRATION_SOURCE,

		/// camera name
		CAMERA_NAME,

		/// line identifier
		SENSOR_LINE,

		/// calibration date
		CALIBRATION_DATE,

		/// IMU misalignment values
		IMU_MISALIGNMENT,

		/// focal length (= principal distance)
		FOCAL_LENGTH_MM,

		/// number of calibrated pixel
		NUM_PIXELS,

		/// pixel size
		PIXEL_SIZE,

		/// start of calibration data
		START_XY,

		/// end of calibration data
		END_XY,

		/// offset of the PAV rotation center from the entrance pupil
		PAV_Z_OFFSET,

		/// multiplier to scale pixel response for balanced colours
		RADIOMETRIC_GAIN,

		/// multiplier to convert pixel response to physical units
		IRRADIANCE_GAIN,
		
		/// leftside gain (?)
		LEFTSIDE_GAIN,
		
		/// end-of-map
		END_OF_MAP
	};

	/// constructor
	ADSCamera();
	
	/**@name set/get member data */
	//@{
		/// get sensor name
		string type_str() const;
	
		/// get keyword name
		string key_str() const;
	//@}
	
	/**@name read/write/print camera parameters
	 *  Read ADS camera calibration data.
	 *  The order of header lines is not fixed.
	 *  The Parameters are identified by kewords.
	 *  For the valid keywords see the map structure definition
	 * in the corresponding source file.
	 */
	//@{
		bool read(istream& is);
	
		/// write
		void write(ostream& os) const;
		
		/// print
		void print(ostream& os) const;
	//@}
	
	/**@name get code.
	 *
	 *  @param name
	 *  @return code
	 */
	static ekey code(const char* name);

private:
	/// version number of the calibration file (string or integer?)
	string		_calib_version;

	/// name of calibration source
	string		_calib_source;

	/// camera name
	string		_camera_name;

	/// sensor line
	string		_sensor_line;

	/// calibration date
	string		_calib_date;

	/// map structure
	struct map_t
	{
		/// code
		ekey	code;

		/// name
		const char *name;
	};

	/// map
	static struct map_t _map[];

	/// strip quotation marks
	string strip(string& s)
	{
		string::size_type b = s.find('\"') + 1;
		string::size_type n = s.rfind('\"') - b;
		return s.substr(b, n);
	};
};

} // namespace DEP

#endif
