/**********************************************************************/
//
// FILE:    adsextori.cpp
//          ADS exterior orientation functions.
// CREATED: 10 Feb 04  (ds)
// CHANGES: see include file
// REFS:    dito
//
// (c) 2004 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#include <strstream>
#include <fstream>
#include "mpc.h"
#include "adsorient.h"

namespace DEP
{

// constructor

ADSOrientation::ADSOrientation() : OrientationData(),
	_version(),
	_major(0),
	_minor(0),
	_datasource(),
	_projectname(),
	_stripname(),
	_numsets(0),
	_posunits(METERS),
	_angularunits(DEGREES),
	_rotationseq(OPK),
	_abstime(0),
	_comments(),
	_basetime(0),
	_basex(0),
	_basey(0),
	_basez(0),
	_pprec(1000),
	_aprec(1000000),
	_tprec(10000),
	_prec_sdev_pos(0),
	_prec_sdev_ang(0)
{
	_transpose = true; // the tranposed form is required!!!
}


// read

bool ADSOrientation::read(const char* fname)
{
	ifstream is(fname, ios_base::binary);
	if (!is) return false;
	
	// read header
	read_header(is);
	
	// number of rotation parameters
	//_num_rot = 3;
	
	double rho = UnitAngle::rho(_uangle);
	//cout << rho << endl;

	// read data
	is.seekg(512, ios::beg);
	ODF odf;
	_data = vector<OrientationDataElement3>(_numsets);
	for (int i = 0; i < _numsets; i++)
		{
		is.read((char*)&odf, sizeof(odf));
		
		_data[i].t = double(odf.t) / double(_tprec) + double(_basetime);
		_data[i].xo = double(odf.xo) / double(_pprec) + double(_basex);
		_data[i].yo = double(odf.yo) / double(_pprec) + double(_basey);
		_data[i].zo = double(odf.zo) / double(_pprec) + double(_basez);
		_data[i].a1 = double(odf.omega) / double(_aprec) / rho;
		_data[i].a2 = double(odf.phi) / double(_aprec) / rho;
		_data[i].a3 = double(odf.kappa) / double(_aprec) / rho;
		}

	return true;
}


// write

void ADSOrientation::write(ostream& os) const
{
	cout << "ERROR: ADSOrientation::write: method still missing!!!" << endl;
	assert(false);
}


// get file format

string ADSOrientation::fmt() const
{
	ostringstream oss;
	oss << _version;
	return oss.str();
}


// read

bool ADSOrientation::read_header(istream& is)
{
	is.read(_version, sizeof(_version));
	//cout << _version << endl;

	// test version number
	char c = 0;
	istrstream iss(_version+3);
	iss >> _major >> c;
	if (c == '.') iss >> _minor;
	//cout << _major << endl;
	//cout << _minor << endl;

	is.seekg(16, ios::beg);
	is.read(_datasource, sizeof(_datasource));

	is.seekg(48, ios::beg);
	is.read(_projectname, sizeof(_projectname));

	is.seekg(112, ios::beg);
	is.read(_stripname, sizeof(_stripname));

	is.seekg(176, ios::beg);
	is >> _numsets;

	is.seekg(192, ios::beg);
	is >> _posunits;

	double sc = 0.;
	switch (_posunits)
		{
		case METERS: sc = 1.; break;

		case FEET: sc = 0.3048; break;
			
		default:
			return false;
			break;
		}

	if (_major >= 1 && _minor >= 1)
		{
		is.seekg(194, ios::beg);
		is >> _pprec;
		}

	is.seekg(208, ios::beg);
	is >> _angularunits;

	switch (_angularunits)
		{
		case RADIANS: _uangle = UnitAngle::RADIAN; break;
		case DEGREES: _uangle = UnitAngle::DEGREE; break;

		default:
			return false;
			break;
		}
	_uangle = UnitAngle::RADIAN; // angular unit is always radian!!!

	if (_major >= 1 && _minor >= 1)
		{
		is.seekg(210, ios::beg);
		is >> _aprec;
		}

	is.seekg(224, ios::beg);
	is >> _rotationseq;

	switch (_rotationseq)
		{
		case OPK: _rtype = ERotMatrix3::ANGLES_OPK_FIX; break;

		default:
		return false;
			break;
		}

	is.seekg(240, ios::beg);
	is >> _abstime;

	is.seekg(256, ios::beg);
	is.read(_comments, sizeof(_comments));

	is.seekg(384, ios::beg);
	is >> _basetime;

	if (_major >= 1 && _minor >= 1)
		{
		is.seekg(396, ios::beg);
		is >> _tprec;
		}

	is.seekg(416, ios::beg);
	is >> _basex;

	if (_major >= 1 && _minor >= 1)
		{
		is.seekg(428, ios::beg);
		is >> _prec_sdev_pos;
		}

	is.seekg(448, ios::beg);
	is >> _basey;

	if (_major >= 1 && _minor >= 1)
		{
		is.seekg(460, ios::beg);
		is >> _prec_sdev_ang;
		}

	is.seekg(480, ios::beg);
	is >> _basez;
	//}
	/*
	else if ()
	{

	}
	else
	{
		return false;
	}
	*/

	return true;
}


// print

void ADSOrientation::print_header(ostream& os) const
{
	Form fix3(3); fix3.fixed().width(12);
	os << "ODF header information\n";
	os << "----------------------\n";
	os << "File ID + Version        :  " << _version << endl;
	os << "Data source              :  " << _datasource << endl;
	os << "Project name             :  " << _projectname << endl;
	os << "Strip name               :  " << _stripname << endl;
	os << "Number of data sets      :  " << _numsets << endl;
	os << "Units position           :  " << _posunits << " (";
	switch (_posunits)
		{
		case METERS: os << "meter"; break;
		case FEET: os << "feet"; break;
		}
	os << ")\n";
	os << "Precision position       :  " << _pprec << endl;
	os << "Unit angles              :  " << _angularunits << " (";
	switch (_angularunits)
		{
		case RADIANS: os << "rad"; break;
		case DEGREES: os << "deg"; break;
		}
	os << ")\n";
	os << "Precision angles         :  " << _aprec << endl;
	os << "Rotation sequence        :  " << _rotationseq << endl;
	os << "Absolute time            :  " << _abstime << endl;
	os << "Comments                 :  " << _comments << endl;
	os << "Base time                :  " << _basetime << endl;
	os << "Precision time           :  " << _tprec << endl;
	os << "Base X                   :  " << _basex << endl;
	os << "Prec. std. dev. position :  " << _prec_sdev_pos << endl;
	os << "Base Y                   :  " << _basey << endl;
	os << "Prec. std. dev. angles   :  " << _prec_sdev_ang << endl;
	os << "Base Z                   :  " << _basez << endl;

	//if (_major == 1 && _minor < 3)
	//	os << "\nATTENTION: angular unit is always radian!!!\n";
}

} // namespace DEP
