/* $Id: apset.cpp,v 1.17 2007/12/05 15:05:42 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "depvec.h"
#include "apset.h"
#include "params.h"
#include "camera.h"
//#include "image.h"
#include "drvsub.h"

namespace DEP
{

// allocate memory for parameters and std. dev.

void APSet::allocate(size_t sz)
{
	_size = sz;
	_ap = new double[_size];
	assert(_ap != 0);
	_sap = new double[_size];
	assert(_sap != 0);
}


// free memory for parameters and std. dev.

void APSet::free()
{
	delete[] _ap;
	delete[] _sap;
}


// new size for parameters and std. dev.

void APSet::newsize(size_t sz)
{
	if (_size == sz) return;
	free();
	allocate(sz);
}


// default constructor

APSet::APSet(size_t sz) :
	AdjObject(),
	_list(),
	_derivs()
{
	allocate(sz);
}


// destructor

APSet::~APSet()
{
	free();
}


// read

bool APSet::read(istream& is)
{
	// read AP set ID
	is >> _id;
	if (!is) return false;
	
	// read camera IDs
	string id;
	while (is >> id && id.compare("stop-dep") != 0)
		_list.push_back(id);

	// read values and std. devs.
	for (size_t i = 0; i < _size; i++)
		is >> _ap[i] >> _sap[i];

	_approx_flag = true;

	return true;
}


// write

void APSet::write(ostream& os) const
{
	Form sci4(4); sci4.scientific().width(12);
	Form sci8(8); sci8.scientific().width(16);

	os << key_str() << "   ";
	os << _id << "   ";

	size_t i = 0;
	for (; i < _list.size(); i++)
		{
		if ((i + 1) % 4 == 0) os << endl;
		os << "  " << _list[i];
		}
	os << "\nstop-dep\n";

	for (i = 0; i < _size; i++)
		os << sci8(_ap[i]) << "   " << sci4(_sap[i]) << endl;
	os << endl;
}


// print

void APSet::print(ostream& os) const
{
	os << "AP set ID            :  " << _id << endl;
	os << "AP set type          :  " << type_str() << endl;
	os << "Number of parameters :  " << _size << endl;
	os << "Number of cameras using this set :  " << _list.size()
		<< endl;
	os << "Cameras     :  " << endl;
	size_t i = 0;
	for (; i < _list.size(); i++)
		{
		if (i % 4 == 0) os << endl;
		os << "  " << _list[i];
		}
	os << endl;
	os << endl;

	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form sci12(3); sci12.scientific().width(12);

	os << "  Param.     State         Value       Std.dev.\n\n";
	for (i = 0; i < _size; i++)
		os << "  " << str8(uname(i)) << "   " << str4(state_str(_sap[i]))
			<< "   " << sci12(_ap[i]) << "   " << sci12(_sap[i]) << endl;

	os << endl;
}



// get adjustment object class name

string APSet::class_str() const
{
	ostringstream oss;
	oss << "AP set";
	return oss.str();
}


// get parameter name

string APSet::uname(size_t i) const
{
	assert(i < _size);

	ostringstream oss;
	oss << "a" << (i + 1);

	return oss.str();
}


// set upari

bool APSet::set_upari()
{
	if (create() != true) return false;

	for (size_t i = 0; i < _size; i++)
		set(i, &_ap[i], &_sap[i]);

	return true;
}


// attach AP set to camera

bool APSet::attach(const vector<Camera*>& cams)
//bool APSet::attach(const vector<Image*>& images)
{
/*
	Camera* cam = find(cams, _camera_id);
	if (cam == 0)
		{
		cout << "ERROR: AP set " << _id
			<< " : camera " << _camera_id << " not found.\n";
		return false;
		}
	cam->attach(this);
	cam->inuse(true);
	_camera = cam;
	inuse(true);
*/
	vector<string>::const_iterator it;
    for (it = _list.begin(); it != _list.end(); ++it)
		{
		Camera* cam = find(cams, *it);
		if (cam == 0)
			{
			cout << "ERROR: Camera " << *it
				<< " for AP set " << _id << " not found.\n";
			return false;
			}
			
		cam->attach(this);
		inuse(true);
		}

	return true;
}


// calculate correction

void APSet::calc_corr(const DPoint3& pho_c, DPoint2& cor) const
{
	//cout << "APSet::calc_corr: " << cor << endl;

	DerivSub derivs;

	get_derivs(pho_c, derivs);

	cor.x = 0.;
	cor.y = 0.;
	for (size_t i = 0; i < _size; i++)
		{
		cor.x += derivs(0, i) * _ap[i];
		cor.y += derivs(1, i) * _ap[i];
		}
}


// calculate correction

void APSet::calc_corr(const string& flgs, const DPoint3& pho_c,
DPoint2& cor) const
{
	assert(flgs.size() == _size);

	DerivSub derivs;

	get_derivs(pho_c, derivs);

	cor.x = 0.;
	cor.y = 0.;
	for (size_t i = 0; i < _size; i++)
		{
		if (flgs[i] == '1')
			{
			cor.x += derivs(0, i) * _ap[i];
			cor.y += derivs(1, i) * _ap[i];
			}
		}
		
	//cout << "APSet::calc_corr: " << cor << endl;
}


// calculate correction

DPoint2 APSet::calc_corr(const DPoint3& pho_c) const
{
	DPoint2 cor;
	calc_corr(pho_c, cor);
	return cor;
}


// update derivatives

void APSet::update_derivs(const DPoint3& pho_c)
{
	get_derivs(pho_c, _derivs);
}


// print influence of APs on image points

void APSet::print_grid(ostream& os, const string& flags,
double xo, double yo, double dx, double dy, int nx, int ny, double c) const
{
	assert(flags.size() == _size);

	Form fix(1); fix.fixed().showpos().width(6);

	DPoint3		pho_c;	// photo coords.
	DPoint2		cor_c;	// correction

	pho_c.z = c;

	pho_c.y = yo;
	for (int j = 0; j < ny; j++, pho_c.y -= dy)
		{
		pho_c.x = xo;
		for (int i = 0; i < nx; i++, pho_c.x += dx)
			{
			calc_corr(flags, pho_c, cor_c);
			cor_c *= 1e6; // convert into um 
			os << " (" << fix(cor_c.x) << "," << fix(cor_c.y) << ")";
			}
		os << endl;
		}
}


// print influence of APs on image points

bool APSet::init_grid(ostream& os, const Params& pr, const Camera& cam,
double& xo, double& yo, double& dx, double& dy,
int& nx, int& ny, double& c) const
{
	os << "AP set ID            :  " << _id << endl;
	os << "AP set type          :  " << type_str() << endl;

	// grid size
	nx = pr.pr_aps_nx;
	ny = pr.pr_aps_ny;

	// intervall
	dx = cam.get_size_x() / double(nx - 1);
	dy = cam.get_size_y() / double(ny - 1);
	//cout << dx << endl;
	//cout << dy << endl;
	
	// top-left origin
	xo = - double(nx - 1) / 2. * dx;
	yo =   double(ny - 1) / 2. * dy;
	//cout << xo << endl;
	//cout << yo << endl;
	
	// range of coords.
	double xmin = xo;
	double xmax = xo + double(nx - 1) * dx;
	double ymin = yo - double(ny - 1) * dy;
	double ymax = yo;

	// focal length
	c = cam.ck();

	Form fix(1); fix.fixed().showpos().width(6);

	os << "Effects on photo-coordinates\n";
	os << "  - range in x       :  min = " << fix(xmin*1e3)
		<< " mm ,  max = " << fix(xmax*1e3) << " mm\n";
	os << "  - range in y       :  min = " << fix(ymin*1e3)
		<< " mm ,  max = " << fix(ymax*1e3) << " mm\n";
	os << endl;

	return true;
}


// print influence of APs on image points

void APSet::print_aps_influence(ostream& os, const Params& pr,
const Camera& cam) const
{
	int		nx, ny;
	double	xo, yo;
	double	dx, dy;
	double	c;
	string	flgs(_size, '1');

	if (init_grid(os, pr, cam, xo, yo, dx, dy, nx, ny, c) != true)
		{
		cout << "ERROR: Initilization of AP grid failed.\n";
		return;
		}

	os << type_str() << ":\n\n";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;
}


// get type name

string APSetIO::type_str() const
{
	ostringstream oss;
	oss << "inner orientation";
	return oss.str();
}


// get key name

string APSetIO::key_str() const
{
	ostringstream oss;
	oss << "inner-or";
	return oss.str();
}


// get parameter name

string APSetIO::uname(size_t i) const
{
	ostringstream oss;

	switch (i)
		{
		case 0: oss << "dxp"; break;
		case 1: oss << "dyp"; break;
		case 2: oss << "dc"; break;
		default:
			assert(false);
			break;
		}

	return oss.str();
}


// get derivatives

void APSetIO::get_derivs(const DPoint3& pho_c, DerivSub& derivs) const
{
	derivs.index(_index);
	derivs.newsize(2, _size);

	double xb = pho_c.x;
	double yb = pho_c.y;
	double zb = pho_c.z;

	derivs(0,0) = 1.;
	derivs(0,1) = 0.;
	derivs(0,2) = - xb / zb;

	derivs(1,0) = 0.;
	derivs(1,1) = 1.;
	derivs(1,2) = - yb / zb;
}


// print influence of APs on image points

void APSetIO::print_aps_influence(ostream& os, const Params& pr,
const Camera& cam) const
{
	int		nx, ny;
	double	xo, yo;
	double	dx, dy;
	double	c;
	string	flgs = "001";

	if (init_grid(os, pr, cam, xo, yo, dx, dy, nx, ny, c) != true)
		{
		cout << "ERROR: Initilization of AP grid failed.\n";
		return;
		}

	os << "inner orientation (dc only):\n\n";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;
}


// update IO parameters

void APSetIO::update(double& xp, double& yp, double& dc)
{
	assert(_size >= 3);

	xp += _ap[0];
	yp += _ap[1];
	dc += _ap[2];

	_ap[0] = _ap[1] = _ap[2] = 0.;
}


// get type name

string APSetGAP::type_str() const
{
	//ostringstream oss("AP GAP");
	ostringstream oss;
	oss << "GAP";
	return oss.str();
}


// get key name

string APSetGAP::key_str() const
{
	//ostringstream oss("ap-gap");
	ostringstream oss;
	oss << "gap";
	return oss.str();
}


// get parameter name

string APSetGAP::uname(size_t i) const
{
	ostringstream oss;

	switch (i)
		{
		case 0: oss << "dxp"; break;
		case 1: oss << "dyp"; break;
		case 2: oss << "dc"; break;
		case 3: oss << "sx"; break;
		case 4: oss << "a1"; break;
		case 5: oss << "K1"; break;
		case 6: oss << "K2"; break;
		case 7: oss << "K3"; break;
		case 8: oss << "P1"; break;
		case 9: oss << "P2"; break;
		default:
			assert(false);
			break;
		}

	return oss.str();
}


// get derivatives

void APSetGAP::get_derivs(const DPoint3& pho_c, DerivSub& derivs) const
{
	derivs.index(_index);
	derivs.newsize(2, _size);

	double xb = pho_c.x;
	double yb = pho_c.y;
	double zb = pho_c.z;
	double xb2 = xb * xb;
	double yb2 = yb * yb;
	double r2 = xb2 + yb2;
	double r4 = r2 * r2;
	double r6 = r4 * r2;

	derivs(0,0) = 1.;
	derivs(0,1) = 0.;
	derivs(0,2) = - xb / zb;
	derivs(0,3) = - xb;
	derivs(0,4) = yb;
	derivs(0,5) = xb * r2;
	derivs(0,6) = xb * r4;
	derivs(0,7) = xb * r6;
	derivs(0,8) = 3. * xb2 + yb2;
	derivs(0,9) = 2. * xb * yb;

	derivs(1,0) = 0.;
	derivs(1,1) = 1.;
	derivs(1,2) = - yb / zb;
	derivs(1,3) = 0.;
	derivs(1,4) = xb;
	derivs(1,5) = yb * r2;
	derivs(1,6) = yb * r4;
	derivs(1,7) = yb * r6;
	derivs(1,8) = 2. * xb * yb;
	derivs(1,9) = 3. * yb2 + xb2;
}


// print influence of APs on image points

void APSetGAP::print_aps_influence(ostream& os, const Params& pr,
const Camera& cam) const
{
	int		nx, ny;
	double	xo, yo;
	double	dx, dy;
	double	c;
	string	flgs;

	if (init_grid(os, pr, cam, xo, yo, dx, dy, nx, ny, c) != true)
		{
		cout << "ERROR: Initilization of AP grid failed.\n";
		return;
		}

	os << "inner orientation (dc only):\n\n";
	flgs = "0010000000";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;

	os << "radial symmetric distortion (K1, K2, K3):\n\n";
	flgs = "0000011100";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;

	os << "decentering distortion (P1, P2):\n\n";
	flgs = "0000000011";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;

	os << "in-plane distortion (sx, a1):\n\n";
	flgs = "0001100000";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;

	os << "total distortion:\n\n";
	flgs = "0011111111";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;
}


// get type name

string APSetAustralis::type_str() const
{
	ostringstream oss;
	oss << "Australis";
	return oss.str();
}


// get key name

string APSetAustralis::key_str() const
{
	ostringstream oss;
	oss << "australis";
	return oss.str();
}


// get parameter name

string APSetAustralis::uname(size_t i) const
{
	ostringstream oss;
	
	switch (i)
		{
		case 0: oss << "dxp"; break;
		case 1: oss << "dyp"; break;
		case 2: oss << "dc"; break;
		case 3: oss << "K1"; break;
		case 4: oss << "K2"; break;
		case 5: oss << "K3"; break;
		case 6: oss << "P1"; break;
		case 7: oss << "P2"; break;
		case 8: oss << "b1"; break;
		case 9: oss << "b2"; break;
		default:
			assert(false);
			break;
		}

	return oss.str();
}


// get derivatives

void APSetAustralis::get_derivs(const DPoint3& pho_c, DerivSub& derivs) const
{
	derivs.index(_index);
	derivs.newsize(2, _size);

	double xb = pho_c.x;
	double yb = pho_c.y;
	double zb = pho_c.z;
	double xb2 = xb * xb;
	double yb2 = yb * yb;
	double r2 = xb2 + yb2;
	double r4 = r2 * r2;
	double r6 = r4 * r2;

	derivs(0,0) = 1.;
	derivs(0,1) = 0.;
	derivs(0,2) = - xb / zb;
	derivs(0,3) = xb * r2;
	derivs(0,4) = xb * r4;
	derivs(0,5) = xb * r6;
	derivs(0,6) = 2. * xb2 + r2;
	derivs(0,7) = 2. * xb * yb;
	derivs(0,8) = xb;
	derivs(0,9) = yb;

	derivs(1,0) = 0.;
	derivs(1,1) = 1.;
	derivs(1,2) = - yb / zb;
	derivs(1,3) = yb * r2;
	derivs(1,4) = yb * r4;
	derivs(1,5) = yb * r6;
	derivs(1,6) = 2. * xb * yb;
	derivs(1,7) = 2. * yb2 + r2;
	derivs(1,8) = 0.;
	derivs(1,9) = 0.;
}


// read

bool APSetNSE::read(istream& is)
{
	// read AP set ID
	is >> _id;
	if (!is) return false;

	// read camera IDs
	string id;
	while (is >> id && id.compare("stop-dep") != 0)
		_list.push_back(id);

	// second zero crossing
	is >> _r0;

	// convert mm into m
	_r0 *= 1e-3;

	// read values and std. devs.
	for (size_t i = 0; i < _size; i++)
		is >> _ap[i] >> _sap[i];

	_approx_flag = true;

	return true;
}


// write

void APSetNSE::write(ostream& os) const
{
	Form fix3(3); fix3.fixed().width(6);
	Form sci4(4); sci4.scientific().width(12);
	Form sci8(8); sci8.scientific().width(16);

	os << key_str() << "   ";
	os << _id << "   ";

	size_t i = 0;
	for (; i < _list.size(); i++)
		{
		if ((i + 1) % 4 == 0) os << endl;
		os << "  " << _list[i];
		}
	os << "\nstop-dep\n";
	
	os << fix3(_r0 * 1e3) << endl;

	for (i = 0; i < _size; i++)
		os << sci8(_ap[i]) << "   " << sci4(_sap[i]) << endl;
	os << endl;
}


// print

void APSetNSE::print(ostream& os) const
{
	Form fix3(3); fix3.fixed().width(6);

	os << "AP set ID   :  " << _id << endl;
	os << "AP set type :  " << type_str() << endl;
	os << "Number of cameras using this set :  " << _list.size()
		<< endl;
	os << "Cameras     :  " << endl;
	size_t i = 0;
	for (; i < _list.size(); i++)
		{
		if ((i + 1) % 4 == 0) os << endl;
		os << "  " << _list[i];
		}
	os << endl;
	os << "2nd zero crossing :  r0 = " << fix3(_r0 * 1e3) << " mm" << endl;
	os << endl;

	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form sci12(3); sci12.scientific().width(12);

	os << "  Param.     State         Value       Std.dev.\n\n";
	for (i = 0; i < _size; i++)
		os << "  " << str8(uname(i)) << "   " << str4(state_str(_sap[i]))
			<< "   " << sci12(_ap[i]) << "   " << sci12(_sap[i]) << endl;

	os << endl;
}


// get type name

string APSetNSE::type_str() const
{
	ostringstream oss;
	oss << "NSE";
	return oss.str();
}


// get key name

string APSetNSE::key_str() const
{
	ostringstream oss;
	oss << "nse";
	return oss.str();
}


// get parameter name

string APSetNSE::uname(size_t i) const
{
	ostringstream oss;

	switch (i)
		{
		case 0: oss << "dxp"; break;
		case 1: oss << "dyp"; break;
		case 2: oss << "dc"; break;
		case 3: oss << "A1"; break;
		case 4: oss << "A2"; break;
		case 5: oss << "A3"; break;
		case 6: oss << "A4"; break;
		case 7: oss << "B1"; break;
		case 8: oss << "B2"; break;
		case 9: oss << "sxy"; break;
		case 10: oss << "C2"; break;
		default:
			assert(false);
			break;
		}

	return oss.str();
}


// get derivatives

void APSetNSE::get_derivs(const DPoint3& pho_c, DerivSub& derivs) const
{
	derivs.index(_index);
	derivs.newsize(2, _size);

	double xb = pho_c.x;
	double yb = pho_c.y;
	double zb = pho_c.z;
	double xb2 = xb * xb;
	double yb2 = yb * yb;
	double r2 = xb2 + yb2;
	double r4 = r2 * r2;
	double r6 = r4 * r2;
	double r8 = r4 * r4;
	double r02 = _r0 * _r0;
	double r04 = r02 * r02;
	double r06 = r04 * r02;
	double r08 = r04 * r04;

	derivs(0,0) = 1.;
	derivs(0,1) = 0.;
	derivs(0,2) = - xb / zb;
	derivs(0,3) = xb * (r2 - r02);
	derivs(0,4) = xb * (r4 - r04);
	derivs(0,5) = xb * (r6 - r06);
	derivs(0,6) = xb * (r8 - r08);
	derivs(0,7) = r2 + 2. * xb2;
	derivs(0,8) = 2. * xb * yb;
	derivs(0,9) = xb;
	derivs(0,10) = yb;
	
	derivs(1,0) = 0.;
	derivs(1,1) = 1.;
	derivs(1,2) = - yb / zb;
	derivs(1,3) = yb * (r2 - r02);
	derivs(1,4) = yb * (r4 - r04);
	derivs(1,5) = yb * (r6 - r06);
	derivs(1,6) = yb * (r8 - r08);
	derivs(1,7) = 2. * xb * yb;
	derivs(1,8) = r2 + 2. * yb2;
	derivs(1,9) = 0.;
	derivs(1,10) = 0.;
}


// print influence of APs on image points

void APSetAustralis::print_aps_influence(ostream& os, const Params& pr,
const Camera& cam) const
{
	int		nx, ny;
	double	xo, yo;
	double	dx, dy;
	double	c;
	string	flgs;

	if (init_grid(os, pr, cam, xo, yo, dx, dy, nx, ny, c) != true)
		{
		cout << "ERROR: Initilization of AP grid failed.\n";
		return;
		}

	os << "inner orientation (dc only):\n\n";
	flgs = "0010000000";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;

	os << "radial symmetric distortion (K1, K2, K3):\n\n";
	flgs = "0001110000";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;

	os << "decentering distortion (P1, P2):\n\n";
	flgs = "0000001100";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;

	os << "in-plane distortion (b1, b2):\n\n";
	flgs = "0000000011";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;

	os << "total distortion:\n\n";
	flgs = "0011111111";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;
}


// read

bool APSetP::read(istream& is)
{
	// read AP set ID
	is >> _id;
	if (!is) return false;

	// read camera IDs
	string id;
	while (is >> id && id.compare("stop-dep") != 0)
		_list.push_back(id);

	// grid width
	is >> _b;

	// convert mm into m
	_b *= 1e-3;

	// read values and std. devs.
	for (size_t i = 0; i < _size; i++)
		is >> _ap[i] >> _sap[i];

	_approx_flag = true;

	return true;
}


// write

void APSetP::write(ostream& os) const
{
	Form fix3(3); fix3.fixed().width(6);
	Form sci4(4); sci4.scientific().width(12);
	Form sci8(8); sci8.scientific().width(16);

	os << key_str() << "   ";
	os << _id << "   ";

	size_t i = 0;
	for (; i < _list.size(); i++)
		{
		if ((i + 1) % 4 == 0) os << endl;
		os << "  " << _list[i];
		}
	os << "\nstop-dep\n";
	
	os << fix3(_b * 1e3) << endl;

	for (i = 0; i < _size; i++)
		os << sci8(_ap[i]) << "   " << sci4(_sap[i]) << endl;
	os << endl;
}


// print

void APSetP::print(ostream& os) const
{
	Form fix3(3); fix3.fixed().width(6);

	os << "AP set ID   :  " << _id << endl;
	os << "AP set type :  " << type_str() << endl;
	os << "Number of cameras using this set :  " << _list.size()
		<< endl;
	os << "Cameras     :  " << endl;
	size_t i = 0;
	for (; i < _list.size(); i++)
		{
		if ((i + 1) % 4 == 0) os << endl;
		os << "  " << _list[i];
		}
	os << endl;
	os << "Grid width  :  b = " << fix3(_b * 1e3) << " mm" << endl;
	os << endl;

	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form sci12(3); sci12.scientific().width(12);

	os << "  Param.     State         Value       Std.dev.\n\n";
	for (i = 0; i < _size; i++)
		os << "  " << str8(uname(i)) << "   " << str4(state_str(_sap[i]))
			<< "   " << sci12(_ap[i]) << "   " << sci12(_sap[i]) << endl;

	os << endl;
}


// get type name

string APSetP12::type_str() const
{
	ostringstream oss;
	oss << "12-parameter (Ebner)";
	return oss.str();
}


// get key name

string APSetP12::key_str() const
{
	ostringstream oss;
	oss << "p12";
	return oss.str();
}


// get derivatives

void APSetP12::get_derivs(const DPoint3& pho_c, DerivSub& derivs) const
{
	derivs.index(_index);
	derivs.newsize(2, _size);

	// image space vector elements
	double xb = pho_c.x;
	double yb = pho_c.y;
	//double zb = pho_c.z unused!!!

	double k = xb * xb - 2. / 3. * _b * _b;
	double l = yb * yb - 2. / 3. * _b * _b;

	derivs(0,0) = xb;
	derivs(0,1) = yb;
	derivs(0,2) = -2. * k;
	derivs(0,3) = xb * yb;
	derivs(0,4) = l;
	derivs(0,5) = 0.;
	derivs(0,6) = xb * l;
	derivs(0,7) = 0.;
	derivs(0,8) = yb * k;
	derivs(0,9) = 0.;
	derivs(0,10) = k * l;
	derivs(0,11) = 0.;

	derivs(1,0) = - yb;
	derivs(1,1) = xb;
	derivs(1,2) = xb * yb;
	derivs(1,3) = -2. * l;
	derivs(1,4) = 0.;
	derivs(1,5) = k;
	derivs(1,6) = 0.;
	derivs(1,7) = yb * k;
	derivs(1,8) = 0.;
	derivs(1,9) = xb * l;
	derivs(1,10) = 0.;
	derivs(1,11) = k * l;
}


// get type name

string APSetP44::type_str() const
{
	ostringstream oss;
	oss << "44-parameter (Gruen)";
	return oss.str();
}


// get key name

string APSetP44::key_str() const
{
	ostringstream oss;
	oss << "p44";
	return oss.str();
}


// get derivatives

void APSetP44::get_derivs(const DPoint3& pho_c, DerivSub& derivs) const
{
	derivs.index(_index);
	derivs.newsize(2, _size);

	// image space vector elements
	double xb = pho_c.x;
	double yb = pho_c.y;
	//double zb = pho_c.z unused!!!

	double xb2 = xb * xb;
	double yb2 = yb * yb;
	double b2 = _b * _b;
	double b4 = b2 * b2;
	double k = xb2 - b2 / 2.;
	double l = yb2 - b2 / 2.;
	double p = xb2 - 17. / 20. * b2;
	double q = yb2 - 17. / 20. * b2;
	double r = xb2 * (xb2 - 31. / 28. * b2) + 9. / 70. * b4;
	double s = yb2 * (yb2 - 31. / 28. * b2) + 9. / 70. * b4;

	derivs(0,0) = xb;
	derivs(0,1) = yb;
	derivs(0,2) = xb * yb;
	derivs(0,3) = l;
	derivs(0,4) = 0.;
	derivs(0,5) = -10. / 7. * k;
	derivs(0,6) = xb * p;
	derivs(0,7) = yb * k;
	derivs(0,8) = xb * l;
	derivs(0,9) = yb * q;
	derivs(0,10) = 0.;
	derivs(0,11) = 0.;
	derivs(0,12) = 0.;
	derivs(0,13) = 0.;
	derivs(0,14) = r;
	derivs(0,15) = xb * yb * p;
	derivs(0,16) = k * l;
	derivs(0,17) = xb * yb * q;
	derivs(0,18) = s;
	derivs(0,19) = 0.;
	derivs(0,20) = 0.;
	derivs(0,21) = 0.;
	derivs(0,22) = 0.;
	derivs(0,23) = 0.;
	derivs(0,24) = yb * r;
	derivs(0,25) = xb * l * p;
	derivs(0,26) = yb * k * q;
	derivs(0,27) = xb * s;
	derivs(0,28) = 0.;
	derivs(0,29) = 0.;
	derivs(0,30) = 0.;
	derivs(0,31) = 0.;
	derivs(0,32) = l * r;
	derivs(0,33) = xb * yb * p * q;
	derivs(0,34) = k * s;
	derivs(0,35) = 0.;
	derivs(0,36) = 0.;
	derivs(0,37) = 0.;
	derivs(0,38) = yb * q * r;
	derivs(0,39) = xb * p * s;
	derivs(0,40) = 0.;
	derivs(0,41) = 0.;
	derivs(0,42) = r * s;
	derivs(0,43) = 0.;

	derivs(1,0) = - yb;
	derivs(1,1) = xb;
	derivs(1,2) = -10. / 7. * l;
	derivs(1,3) = 0.;
	derivs(1,4) = k;
	derivs(1,5) = xb * yb;
	derivs(1,6) = 0.;
	derivs(1,7) = 0.;
	derivs(1,8) = 0.;
	derivs(1,9) = 0.;
	derivs(1,10) = xb * p;
	derivs(1,11) = yb * k;
	derivs(1,12) = xb * l;
	derivs(1,13) = yb * q;
	derivs(1,14) = 0.;
	derivs(1,15) = 0.;
	derivs(1,16) = 0.;
	derivs(1,17) = 0.;
	derivs(1,18) = 0.;
	derivs(1,19) = r;
	derivs(1,20) = xb * yb * p;
	derivs(1,21) = k * l;
	derivs(1,22) = xb * yb * q;
	derivs(1,23) = s;
	derivs(1,24) = 0.;
	derivs(1,25) = 0.;
	derivs(1,26) = 0.;
	derivs(1,27) = 0.;
	derivs(1,28) = yb * r;
	derivs(1,29) = xb * l * p;
	derivs(1,30) = yb * k * q;
	derivs(1,31) = xb * s;
	derivs(1,32) = 0.;
	derivs(1,33) = 0.;
	derivs(1,34) = 0.;
	derivs(1,35) = l * r;
	derivs(1,36) = xb * yb * p * q;
	derivs(1,37) = k * s;
	derivs(1,38) = 0.;
	derivs(1,39) = 0.;
	derivs(1,40) = yb * q * r;
	derivs(1,41) = xb * p * s;
	derivs(1,42) = 0.;
	derivs(1,43) = r * s;
}


// get type name

string APSetRadial::type_str() const
{
	ostringstream oss;
	oss << "radial distortion";
	return oss.str();
}


// get key name

string APSetRadial::key_str() const
{
	ostringstream oss;
	oss << "radial-dist";
	return oss.str();
}


// get parameter name

string APSetRadial::uname(size_t i) const
{
	ostringstream oss;

	switch (i)
		{
		case 0: oss << "K1"; break;
		case 1: oss << "K2"; break;
		case 2: oss << "K3"; break;
		default:
			assert(false);
			break;
		}

	return oss.str();
}


// get derivatives

void APSetRadial::get_derivs(const DPoint3& pho_c, DerivSub& derivs) const
{
	derivs.index(_index);
	derivs.newsize(2, _size);

	double xb = pho_c.x;
	double yb = pho_c.y;
	double xb2 = xb * xb;
	double yb2 = yb * yb;
	double r2 = xb2 + yb2;
	double r4 = r2 * r2;
	double r6 = r4 * r2;

	derivs(0,0) = xb * r2;
	derivs(0,1) = xb * r4;
	derivs(0,2) = xb * r6;

	derivs(1,0) = yb * r2;
	derivs(1,1) = yb * r4;
	derivs(1,2) = yb * r6;
}


// get type name

string APSetDecentering::type_str() const
{
	ostringstream oss;
	oss << "decentering distortion";
	return oss.str();
}


// get key name

string APSetDecentering::key_str() const
{
	ostringstream oss;
	oss << "decentering-dist";
	return oss.str();
}


// get parameter name

string APSetDecentering::uname(size_t i) const
{
	ostringstream oss;

	switch (i)
		{
		case 0: oss << "P1"; break;
		case 1: oss << "P2"; break;
		default:
			assert(false);
			break;
		}

	return oss.str();
}


// get derivatives

void APSetDecentering::get_derivs(const DPoint3& pho_c, DerivSub& derivs) const
{
	derivs.index(_index);
	derivs.newsize(2, _size);

	double xb = pho_c.x;
	double yb = pho_c.y;
	double xb2 = xb * xb;
	double yb2 = yb * yb;

	derivs(0,0) = 3. * xb2 + yb2;
	derivs(0,1) = 2. * xb * yb;

	derivs(1,0) = 2. * xb * yb;
	derivs(1,1) = 3. * yb2 + xb2;
}


// get type name

string APSetInPlane::type_str() const
{
	ostringstream oss;
	oss << "in-plane distortion";
	return oss.str();
}


// get key name

string APSetInPlane::key_str() const
{
	ostringstream oss;
	oss << "in-plane-dist";
	return oss.str();
}


// get derivatives

void APSetInPlane::get_derivs(const DPoint3& pho_c, DerivSub& derivs) const
{
	derivs.index(_index);
	derivs.newsize(2, _size);

	double xb = pho_c.x;
	double yb = pho_c.y;
	double xb2 = xb * xb;
	double yb2 = yb * yb;

	derivs(0, 0) = xb;
	derivs(0, 1) = yb;
	derivs(0, 2) = xb * yb;
	derivs(0, 3) = yb2;
	derivs(0, 4) = xb2 * yb;
	derivs(0, 5) = xb * yb2;
	derivs(0, 6) = xb2 * yb2;
	derivs(0, 7) = 0.;
	derivs(0, 8) = 0.;
	derivs(0, 9) = 0.;
	derivs(0,10) = 0.;
	derivs(0,11) = 0.;

	derivs(1, 0) = 0.;
	derivs(1, 1) = 0.;
	derivs(1, 2) = 0.;
	derivs(1, 3) = 0.;
	derivs(1, 4) = 0.;
	derivs(1, 5) = 0.;
	derivs(1, 6) = 0.;
	derivs(1, 7) = xb * yb;
	derivs(1, 8) = xb2;
	derivs(1, 9) = xb2 * yb;
	derivs(1,10) = xb * yb2;
	derivs(1,11) = xb2 * yb2;
}


// get type name

string APSetOutOfPlane::type_str() const
{
	ostringstream oss;
	oss << "out-of-plane distortion";
	return oss.str();
}


// get key name

string APSetOutOfPlane::key_str() const
{
	ostringstream oss;
	oss << "out-of-plane-dist";
	return oss.str();
}


// get parameter name

string APSetOutOfPlane::uname(size_t i) const
{
	ostringstream oss;

	switch (i)
		{
		case 0: oss << "a13"; break;
		case 1: oss << "a14"; break;
		case 2: oss << "a15"; break;
		default:
			assert(false);
			break;
		}

	return oss.str();
}


// get derivatives

void APSetOutOfPlane::get_derivs(const DPoint3& pho_c, DerivSub& derivs) const
{
	derivs.index(_index);
	derivs.newsize(2, _size);

	double xb = pho_c.x;
	double yb = pho_c.y;
	double zb = pho_c.z;
	double xb2 = xb * xb;
	double yb2 = yb * yb;
	double xb4 = xb2 * xb2;
	double yb4 = yb2 * yb2;
	double sx = - xb / zb;
	double sy = - yb / zb;

	derivs(0,0) = (xb2 - yb2) * sx;
	derivs(0,1) = (xb2 * yb2) * sx;
	derivs(0,2) = (xb4 - yb4) * sx;

	derivs(1,0) = (xb2 - yb2) * sy;
	derivs(1,1) = (xb2 * yb2) * sy;
	derivs(1,2) = (xb4 - yb4) * sy;
}


// get type name

string APSetBrownEqn22::type_str() const
{
	ostringstream oss;
	oss << "Brown Eqn. 22";
	return oss.str();
}


// get key name

string APSetBrownEqn22::key_str() const
{
	ostringstream oss;
	oss << "brown-eqn-22";
	return oss.str();
}


// get parameter name

string APSetBrownEqn22::uname(size_t i) const
{
	ostringstream oss;

	switch (i)
		{
		case 0: oss << "dxp"; break;
		case 1: oss << "dyp"; break;
		case 2: oss << "dc"; break;
		case 3: oss << "K1"; break;
		case 4: oss << "K2"; break;
		case 5: oss << "K3"; break;
		case 6: oss << "P1"; break;
		case 7: oss << "P2"; break;
		case 8: oss << "a1"; break;
		case 9: oss << "a2"; break;
		case 10: oss << "a3"; break;
		case 11: oss << "a4"; break;
		case 12: oss << "a5"; break;
		case 13: oss << "a6"; break;
		case 14: oss << "a7"; break;
		case 15: oss << "b1"; break;
		case 16: oss << "b2"; break;
		case 17: oss << "b3"; break;
		case 18: oss << "b4"; break;
		case 19: oss << "b5"; break;
		case 20: oss << "b6"; break;
		case 21: oss << "b7"; break;
		case 22: oss << "c1"; break;
		case 23: oss << "c2"; break;
		case 24: oss << "c3"; break;
		case 25: oss << "c4"; break;
		case 26: oss << "c5"; break;
		case 27: oss << "c6"; break;
		case 28: oss << "c7"; break;
		default:
			assert(false);
			break;
		}

	return oss.str();
}


// get derivatives

void APSetBrownEqn22::get_derivs(const DPoint3& pho_c, DerivSub& derivs) const
{
	derivs.index(_index);
	derivs.newsize(2, _size);

	double xb = pho_c.x;
	double yb = pho_c.y;
	double zb = pho_c.z;
	double xb2 = xb * xb;
	double yb2 = yb * yb;
	double r2 = xb2 + yb2;
	double r4 = r2 * r2;
	double r6 = r4 * r2;
	double sx = - xb / zb;
	double sy = - yb / zb;

	derivs(0, 0) = 1.;
	derivs(0, 1) = 0.;
	derivs(0, 2) = sx;
	
	derivs(0, 3) = xb * r2;
	derivs(0, 4) = xb * r4;
	derivs(0, 5) = xb * r6;
	
	derivs(0, 6) = yb2 + 3. * xb2;
	derivs(0, 7) = 2. * xb * yb;
	
	derivs(0, 8) = xb;
	derivs(0, 9) = yb;
	derivs(0,10) = xb2;
	derivs(0,11) = xb * yb;
	derivs(0,12) = yb2;
	derivs(0,13) = xb2 * yb;
	derivs(0,14) = xb * yb2;
	
	derivs(0,15) = 0.;
	derivs(0,16) = 0.;
	derivs(0,17) = 0.;
	derivs(0,18) = 0.;
	derivs(0,19) = 0.;
	derivs(0,20) = 0.;
	derivs(0,21) = 0.;
	
	derivs(0,22) = sx * xb2;
	derivs(0,23) = sx * xb * yb;
	derivs(0,24) = sx * yb2;
	derivs(0,25) = sx * xb * xb2;
	derivs(0,26) = sx * xb2 * yb;
	derivs(0,27) = sx * xb * yb2;
	derivs(0,28) = sx * yb * yb2;
	
	derivs(1, 0) = 0.;
	derivs(1, 1) = 1.;
	derivs(1, 2) = sy;
	
	derivs(1, 3) = yb * r2;
	derivs(1, 4) = yb * r4;
	derivs(1, 5) = yb * r6;
	
	derivs(1, 6) = 2. * xb * yb;
	derivs(1, 7) = xb2 + 3. * yb2;
	
	derivs(1, 8) = 0.;
	derivs(1, 9) = 0.;
	derivs(1,10) = 0.;
	derivs(1,11) = 0.;
	derivs(1,12) = 0.;
	derivs(1,13) = 0.;
	derivs(1,14) = 0.;
	
	derivs(1,15) = xb;
	derivs(1,16) = yb;
	derivs(1,17) = xb2;
	derivs(1,18) = xb * yb;
	derivs(1,19) = yb2;
	derivs(1,20) = xb2 * yb;
	derivs(1,21) = xb * yb2;
	
	derivs(1,22) = sy * xb2;
	derivs(1,23) = sy * xb * yb;
	derivs(1,24) = sy * yb2;
	derivs(1,25) = sy * xb * xb2;
	derivs(1,26) = sy * xb2 * yb;
	derivs(1,27) = sy * xb * yb2;
	derivs(1,28) = sy * yb * yb2;
}

/*
// print influence of APs on image points

void APSetGAP::print_aps_influence(ostream& os, const Params& pr,
const Camera& cam) const
{
	int		nx, ny;
	double	xo, yo;
	double	dx, dy;
	double	c;
	string	flgs;

	if (init_grid(os, pr, cam, xo, yo, dx, dy, nx, ny, c) != true)
		{
		cout << "ERROR: Initilization of AP grid failed.\n";
		return;
		}

	os << "inner orientation (dc only):\n\n";
	flgs = "0010000000";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;

	os << "radial symmetric distortion (K1, K2, K3):\n\n";
	flgs = "0000011100";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;

	os << "decentering distortion (P1, P2):\n\n";
	flgs = "0000000011";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;

	os << "in-plane distortion (sx, a1):\n\n";
	flgs = "0001100000";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;

	os << "total distortion:\n\n";
	flgs = "0011111111";
	print_grid(os, flgs, xo, yo, dx, dy, nx, ny, c);
	os << endl;
}
*/

// get type name

string APSetBrownEqn23::type_str() const
{
	ostringstream oss;
	oss << "Brown Eqn. 23";
	return oss.str();
}


// get key name

string APSetBrownEqn23::key_str() const
{
	ostringstream oss;
	oss << "brown-eqn-23";
	return oss.str();
}


// get parameter name

string APSetBrownEqn23::uname(size_t i) const
{
	ostringstream oss;

	switch (i)
		{
		case 0: oss << "dxp"; break;
		case 1: oss << "dyp"; break;
		case 2: oss << "dc"; break;
		case 3: oss << "K1"; break;
		case 4: oss << "K2"; break;
		case 5: oss << "K3"; break;
		case 6: oss << "a1"; break;
		case 7: oss << "a2"; break;
		case 8: oss << "a3"; break;
		case 9: oss << "a4"; break;
		case 10: oss << "a5"; break;
		case 11: oss << "a6"; break;
		case 12: oss << "a7"; break;
		case 13: oss << "a8"; break;
		case 14: oss << "a9"; break;
		case 15: oss << "a10"; break;
		case 16: oss << "a11"; break;
		case 17: oss << "a12"; break;
		case 18: oss << "a13"; break;
		case 19: oss << "a14"; break;
		case 20: oss << "a15"; break;
		default:
			assert(false);
			break;
		}

	return oss.str();
}


// get derivatives

void APSetBrownEqn23::get_derivs(const DPoint3& pho_c, DerivSub& derivs) const
{
	derivs.index(_index);
	derivs.newsize(2, _size);

	double xb = pho_c.x;
	double yb = pho_c.y;
	double zb = pho_c.z;
	double xb2 = xb * xb;
	double yb2 = yb * yb;
	double xb4 = xb2 * xb2;
	double yb4 = yb2 * yb2;
	double r2 = xb2 + yb2;
	double r4 = r2 * r2;
	double r6 = r4 * r2;
	double sx = - xb / zb;
	double sy = - yb / zb;

	derivs(0, 0) = 1.;
	derivs(0, 1) = 0.;
	derivs(0, 2) = sx;
	
	derivs(0, 3) = xb * r2;
	derivs(0, 4) = xb * r4;
	derivs(0, 5) = xb * r6;
	
	derivs(0, 6) = xb;
	derivs(0, 7) = yb;
	derivs(0, 8) = xb * yb;
	derivs(0, 9) = yb2;
	derivs(0,10) = xb2 * yb;
	derivs(0,11) = xb * yb2;
	derivs(0,12) = xb2 * yb2;
	derivs(0,13) = 0.;
	derivs(0,14) = 0.;
	derivs(0,15) = 0.;
	derivs(0,16) = 0.;
	derivs(0,17) = 0.;
	
	derivs(0,18) = sx * (xb2 - yb2);
	derivs(0,19) = sx * (xb2 * yb2);
	derivs(0,20) = sx * (xb4 - yb4);
	
	derivs(1, 0) = 0.;
	derivs(1, 1) = 1.;
	derivs(1, 2) = sy;
	
	derivs(1, 3) = yb * r2;
	derivs(1, 4) = yb * r4;
	derivs(1, 5) = yb * r6;
	
	derivs(1, 6) = 0.;
	derivs(1, 7) = 0.;
	derivs(1, 8) = 0.;
	derivs(1, 9) = 0.;
	derivs(1,10) = 0.;
	derivs(1,11) = 0.;
	derivs(1,12) = 0.;
	
	derivs(1,13) = xb * yb;
	derivs(1,14) = xb2;
	derivs(1,15) = xb2 * yb;
	derivs(1,16) = xb * yb2;
	derivs(1,17) = xb2 * yb2;
	
	derivs(1,18) = sy * (xb2 - yb2);
	derivs(1,19) = sy * (xb2 * yb2);
	derivs(1,20) = sy * (xb4 - yb4);

}

// non-member functions ...

} // namespace DEP
