/* $Id: attcor.cpp,v 1.15 2007/10/29 11:18:27 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "depvec.h"
#include "attcor.h"
#include "drvsub.h"
#include "image.h"

namespace DEP
{
// constructor

AttCorrection::AttCorrection() : AdjObject(),
_image_ids(),
_a(0), _sa(0), _t0(0.)
{
	// empty
}


// read

bool AttCorrection::read(istream& is)
{
	// read image IDs
	string image_id;
	while (is >> image_id && image_id.compare("stop-dep") != 0)
		_image_ids.push_back(image_id);
	
	// read number of coeffs.
	int sz = 0;
	is >> sz;
	_a = vector<DPoint3>(sz);
	_sa = vector<DPoint3>(sz);

	_size = 3 * _a.size();

	// read coeff. and std. devs.
	for (int i = 0; i < int(_a.size()); i++)
		{
		is >> _a[i].x >> _sa[i].x;
		is >> _a[i].y >> _sa[i].y;
		is >> _a[i].z >> _sa[i].z;
		}

	_approx_flag = true;

	return true;
}


// write
/*
void AttCorrection::write(ostream& os, SDevLim& sdev_lim)
{
	assert(_a.size() == _sa.size());

	os << id << "   " << name << endl;

	for (int i = 0; i < _a.size(); i++)
		{
		os << form("   %10.4e   %10.4e\n", _a[i].x, _sa[i].x);
		os << form("   %10.4e   %10.4e\n", _a[i].y, _sa[i].y);
		os << form("   %10.4e   %10.4e\n", _a[i].z, _sa[i].z);
		os << endl;
		}

	os << endl;
}
*/


// print

void AttCorrection::print(ostream& os) const
{
	assert(_a.size() == _sa.size());

	double rho = AdjObject::lspars.rho();

	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix12(6); fix12.fixed().width(12);
	Form sci12(3); sci12.scientific().width(12);

	os << "Att. corr. ID         :  " << _id << endl;
	os << "Images using this set :  " << _image_ids.size();
	{
		for (int i = 0; i < int(_image_ids.size()); i++)
			{
			if (i % 4 == 0) os << endl;
			os << "  " << _image_ids[i];
			}
		os << endl;
		os << "Reference time     :  t0 = " << fix12(_t0) << " s\n";
		os << "Units of coeffs.   :  " << AdjObject::lspars.u_angle_cstr()
			<< "/s**i\n";
		os << endl;
	}

	os << "  Param.     State         Value       Std.dev.\n\n";
	for (size_t i = 0, k = 0; i < _a.size(); i++)
		{
		os << "  " << str8(uname(k++)) << "   " << str4(state_str(_sa[i].x))
			<< "   " << sci12(_a[i].x * rho) << "   ";
		if (is_rsdev(_sa[i].x))
			os << sci12(_sa[i].x * rho);
		else
			os << sci12(_sa[i].x);
		os << endl;

		os << "  " << str8(uname(k++)) << "   " << str4(state_str(_sa[i].y))
			<< "   " << sci12(_a[i].y * rho) << "   ";
		if (is_rsdev(_sa[i].y))
			os << sci12(_sa[i].y * rho);
		else
			os << sci12(_sa[i].y);
		os << endl;

		os << "  " << str8(uname(k++)) << "   " << str4(state_str(_sa[i].z))
			<< "   " << sci12(_a[i].z * rho) << "   ";
		if (is_rsdev(_sa[i].z))
			os << sci12(_sa[i].z * rho);
		else
			os << sci12(_sa[i].z);
		os << endl << endl;
		}
	
	os << endl;
}


// get adjustment object class name

string AttCorrection::class_str() const
{
	ostringstream oss;
	oss << "attitude correction";
	return oss.str();
}


// get key name

string AttCorrection::key_str() const
{
	ostringstream oss;
	oss << "att-corr";
	return oss.str();
}


// get parameter name

string AttCorrection::uname(size_t i) const
{
	ostringstream oss;
	
	assert(i < 3 * _a.size());
	oss << "c[";
	switch (i % 3)
		{
		case 0: oss << "a1"; break;
		case 1: oss << "a2"; break;
		case 2: oss << "a3"; break;
		}
	oss << "," << (i / 3) << "]";

	return oss.str();
}


// set upari

bool AttCorrection::set_upari()
{
	if (create() != true) return false;

	int k = 0;
	for (size_t i = 0; i < _a.size(); i++)
		{
		set(k, &_a[i].x, &_sa[i].x); k++;
		set(k, &_a[i].y, &_sa[i].y); k++;
		set(k, &_a[i].z, &_sa[i].z); k++;
		}

	return true;
}


// print changes of unknowns without header

void AttCorrection::print_changes(ostream& os, const Vector<double>& dx) const
{
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form sci12(3); sci12.scientific().width(12);

	os << "ID       :  " << _id << endl;
	os << "Index    :  " << _index << endl;
	os << endl;
	
	const double rho = AdjObject::lspars.rho();
	
	os << "  Param.     State"
		"       Approx.         Change       Adjusted\n\n";
		
	int k = _index;
	for (size_t i = 0; i < _size; i++, k++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(i)) << "   "
			<< sci12(par(i) * rho) << "   "
			<< sci12(dx(k) * rho) << "   "
			<< sci12((par(i) + dx(k)) * rho) << endl;
		if ((i + 1) % 3 == 0) os << endl;
		}
	os << endl;
}


// print unknown parameters

void AttCorrection::print_unknowns(ostream& os) const
{
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form sci12(3); sci12.scientific().width(12);

	os << "ID       :  " << _id << endl;
	os << "Index    :  " << _index << endl;
	os << endl;
	
	const double rho = AdjObject::lspars.rho();
	
	os << "  Param.    State      Initial      Adjusted        Change    "
		"Std.dev.\n\n";

	for (size_t i = 0; i < _size; i++)
		{
		os << "  " << str8(uname(i)) << "  "
			<< str4(state_str(i)) << "  "
			<< sci12(ival(i) * rho) << "  "
			<< sci12(par(i) * rho) << "  "
			<< sci12(change(i) * rho);
		if (is_rsdev(sdev(i)))
			os << sci12(sdev(i) * rho);
		else
			os << sci12(sdev(i));
		os << endl;
		if ((i + 1) % 3 == 0) os << endl;
		}
	os << endl;
}


// print precision of unknowns

void AttCorrection::print_precision(ostream& os) const
{
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix3(1); fix3.fixed().width(6);
	Form sci12(3); sci12.scientific().width(12);
	
	os << "ID       :  " << _id << endl;
	os << "Index    :  " << _index << endl;
	os << endl;
	
	const double rho = AdjObject::lspars.rho();
	
	os << "  Param.    State      Initial      Adjusted        Change      "
		"Std.dev.    Ratio\n\n";
	
	for (size_t i = 0; i < _size; i++)
		{
		
		double t_abs = ratio(i); // test value

		os << "  " << str8(uname(i)) << "  "
			<< str4(state_str(i)) << "  "
			<< sci12(ival(i) * rho) << "  "
			<< sci12(par(i) * rho) << "  "
			<< sci12(change(i) * rho) << "  ";
		if (is_rsdev(sdev(i)))
			os << sci12(sdev(i) * rho);
		else
			os << sci12(sdev(i));
		os << "  " << fix3(t_abs);
		
		// test on significance
		if (state(i) == eli)
			os << "  is eliminated";
		else
			{
			if (chk_sign(t_abs))
				os << "  is significant";
			else
				os << "  not significant";
			}
		os << endl;
		if ((i + 1) % 3 == 0) os << endl;
		}
	os << endl;
}


// get position correction
// evaluate polynomial p(dt) with the coefficients a[0..n]

DPoint3 AttCorrection::calc_corr(double t) const
{
	double dt = t - _t0;
	DPoint3 cor = _a[_a.size() - 1];

	for (int j = _a.size() - 2; j >= 0; j--)
		cor = cor * dt + _a[j];
	
	return cor;
}


// attach to images

bool AttCorrection::attach(vector<Image*>& images)
{
	vector<string>::const_iterator it;
	for (it = _image_ids.begin(); it != _image_ids.end(); ++it)
		{
		Image* p = find(images, *it);
		if (p == 0)
			{
			cout << "WARNING: Image '" << *it
				<< "' for att. correction '" << _id << "' not found.\n";
			continue;
		}
		
		p->attach(this);
		this->inuse(true);
		}

	return true;
}


// calculate derivatives

void AttCorrection::calc_derivs(double t, const DerivSub& deriv_rot,
DerivSub& deriv_ac) const
{
	//assert(deriv_eo.rows() == 2);
	//assert(deriv_eo.cols() == 6);

	int sz = 3 * _a.size();
	deriv_ac.index(_index);	// set index to normal equations
	deriv_ac.newsize(2, sz);

	deriv_ac(0,0) = deriv_rot(0,0);
	deriv_ac(1,0) = deriv_rot(1,0);
	deriv_ac(0,1) = deriv_rot(0,1);
	deriv_ac(1,1) = deriv_rot(1,1);
	deriv_ac(0,2) = deriv_rot(0,2);
	deriv_ac(1,2) = deriv_rot(1,2);

	double dt = t - _t0;

	for (int i = 3; i < sz; i++)
		{
		deriv_ac(0,i) = deriv_ac(0,i-3) * dt;
		deriv_ac(1,i) = deriv_ac(1,i-3) * dt;
		}
}

} // namespace DEP
