/* $Id: ccdcam.h,v 1.14 2007/10/11 12:43:57 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_CCDCAM_INCLUDED
#define DEP_CCDCAM_INCLUDED

#include "camera.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{

/** CCD frame camera class.
 *  Special class for CCD frame cameras.
 *  Additionaly to the base class this class contains parameters
 *  of the CCD sensor:
 *  \begin{itemize}
 *  \item number of sensor elements (pixel)
 *  \item pixel spacing
 *  \item location of the Image Coordinate System (ICS)
 *  \end{itemize}
 *  The location of the ICS (in pixel coordinates)
 *  is defined in the centre of the CCD array.
 *
 * @author Dirk Stallmann
 */
class CCDCamera : public Camera
{
public:
	/// constructor
	CCDCamera();
	
	/**@name set/get member data */
	//@{
		/// get camera type
		virtual string type_str() const;
	
		/// get keyword name
		virtual string key_str() const;
		
		/// get type of coordinate system type
		virtual string csys_str() const;
		
		/// get type of coordinate system
		virtual CSystem csystem() const { return PIXEL; }
		
		/// get orientation type
		virtual string orientation_str() const;
		
		/// get number of sensor elements in column direction
		size_t ncols() const { return _nc; }
	
		/// get number of sensor elements in row direction
		size_t nrows() const { return _nr; }
	//@}
	
	/**@name read/write/print camera parameters
	 */
	//@{
		/// read camera parameters
		bool read(istream& is);
	
		/// write camera parameters
		void write(ostream& os) const;
	
		/// print camera parameters
		void print(ostream& os) const;
	//@}

	/**@name sensor/image coordinates transformation
	 *  To convert between sensor (pixel) and image coordinates.
	 */
	//@{
		/// sensor-to-image coords. transformation
		virtual void sensor_to_image(const DPoint2& sen_c,
			DPoint2& img_c) const;

		/// image-to-sensor-coords. transformation
		virtual void image_to_sensor(const DPoint2& img_c,
			DPoint2& sen_c) const;
	//@}

	/** transformation of standard deviations
	 *  from sensor to image coordinates
	 */
	virtual void sdev_sensor_to_image(const DPoint2& sen_s,
		DPoint2& img_s) const;

	/**@name read/write/print sensor coordinates.
	 */
	//@{
		/// read sensor coords.
		virtual bool read_sensor_coords(istream& is,
			vector<ImagePoint*>& impts, bool sdev_flg, const DPoint2& sdev) const;

		/// write sensor coords.
		virtual void write_sensor_coords(ostream& os,
			const vector<ImagePoint*>& impts) const;

		/// print sensor coords.
		virtual void print_sensor_coords(ostream& os,
			const vector<ImagePoint*>& impts, bool ic_out) const;
	//@}

	/// calculate GSD
	void GSD(double hg, double& gsd_c, double& gsd_r);

protected:
	/// number of sensor elements in column direction
	size_t		_nc;

	/// number of sensor elements in row direction
	size_t		_nr;

	/// pixel spacing in column direction
	double		_sc;

	/// pixel spacing in row direction
	double		_sr;

	/// location of the ICS in column direction
	double		_c0;

	/// location of the ICS in row direction
	double		_r0;
};


// non-member functions ...

} // namespace DEP

#endif
