/**********************************************************************/
//
// FILE:    pbcamcal.cpp
//          Calibrated pushbroom camera class functions.
// CREATED: 19 Feb 04  (ds)
// CHANGES: see include file
// REFS:    dito 
//
// (c) 2005 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#include <fstream>
#include "pbcamcal.h"

namespace DEP
{

// constructor

PBCameraCalib::PBCameraCalib() : CCDCamera(),
	_xcal(0),
	_ycal(0)
{
	_nr = 1; // only one row!!!
}


// destructor

PBCameraCalib::~PBCameraCalib()
{
	delete[] _xcal;
	delete[] _ycal;
}


// get string of camera type

string PBCameraCalib::type_str() const
{
	ostringstream oss;
	oss << "pushbroom calibrated";
	return oss.str();
}


// get keyword name

string PBCameraCalib::key_str() const
{
	ostringstream oss;
	oss << "pushbroom-camera-calib";
	return oss.str();
}


// read camera parameter

bool PBCameraCalib::read(istream& is)
{
	// identifier
	is >> _id;
	if (!is) return false;

	// comment
	is >> _comment;
	if (!is) return false;

	// principal distance in mm
	is >> _princ_dist;
	if (!is) return false;

	// convert from mm to m
	_princ_dist *= 1e-3;

	// number of pixel in line and pixel size
	is >> _nc >> _sc >> _sr;
	if (!is) return false;

	// convert from um/pixel to m/pixel
	_sc *= 1e-6;
	_sr *= 1e-6;

	// allocate memory for calibrated pixel positions
	if (_xcal != 0) delete[] _xcal;
	if (_ycal != 0) delete[] _ycal;

	_xcal = new double[_nc];
	_ycal = new double[_nc];

	if (_xcal == 0 || _ycal == 0)
		{
		cout << "ERROR: Memory allocation for pixel data failed.\n";
		return false;
		}

	// read calibrated pixel positions
	size_t num_pix = 0;
	for (size_t i = 0; i < _nc; i++)
		{
		is >> _xcal[i] >> _ycal[i];
		_xcal[i] *= 1e-3; // convert from mm to m
		_ycal[i] *= 1e-3;
		num_pix++;
		}

	// check number of pixels
	if (num_pix != _nc)
		{
		cout << "ERROR: Check number of pixels failed.\n";
		return false;
		}
	
	// calc. ICS location
	_c0 = (_nc - 1) / 2.;
	//_r0 = (_nr - 1) / 2.;

	// calc. image format
	_size_x = _nr * _sr;
	_size_y = _nc * _sc;

	return true;
}


// write camera data

void PBCameraCalib::write(ostream& os) const
{
	assert(false);
}


// print camera parameter

void PBCameraCalib::print(ostream& os) const
{
	CCDCamera::print(os);

	Form int5; int5.width(5);
	Form fix4(4); fix4.fixed().width(10);
	os << "Calibrated pixel positions" << endl;
	size_t k = 0;
	os << "  pixel [" << int5(k) << "] "
		<< "  x = " << fix4(_xcal[k] * 1000.) << " mm"
		<< "  y = " << fix4(_ycal[k] * 1000.) << " mm" << endl;
	k = _nc - 1;
	os << "  pixel [" << int5(k) << "] "
		<< "  x = " << fix4(_xcal[k] * 1000.) << " mm"
		<< "  y = " << fix4(_ycal[k] * 1000.) << " mm" << endl;
	os << endl;
}



// sensor-to-image coordinates transformation

void PBCameraCalib::sensor_to_image(const DPoint2& pix_c,
DPoint2& img_c) const
{
	// the y-component (r) is unused!!!
	int j = int(pix_c.x);
	if (j < 0)
		j = 0; // use the first pixel
	else if (j >= int(_nc) - 1)
		j = int(_nc) - 2; // use the second last pixel

	double a = pix_c.x - double(j);

	double y0 = _xcal[j];
	double y1 = _xcal[j+1];
	img_c.x = (y1 - y0) * a + y0;

	y0 = _ycal[j];
	y1 = _ycal[j+1];
	img_c.y = (y1 - y0) * a + y0;
}


// image-to-pixel image coordinates transformation

void PBCameraCalib::image_to_sensor(const DPoint2& img_c,
DPoint2& pix_c) const
{
	int j = locate(_ycal, img_c.y);
	if (j < 0)
		j = 0; // use the first pixel
	else if (j >= int(_nc) - 1)
		j = int(_nc) - 2; // use the second last pixel

	double y0 = _ycal[j];
	double y1 = _ycal[j+1];

	double a = img_c.y - y0;
	double dy = y1 - y0;

	pix_c.x = double(j) + a / dy;
	// the y-component is unchanged!!!
}


// locate in ordered table

int PBCameraCalib::locate(double xx[], double x) const
{
	//cout << x << endl;
	int jl = -1;	// initilize lower limit
	int ju = _nc;	// and upper limit
	int ascnd = (xx[_nc-1] > xx[0]);
	while (ju - jl > 1)
		{
		int jm = (ju + jl) >> 1;
		//cout << "[" << jm << "] " << absc[jm] << endl;
		if (x > xx[jm] == ascnd)
			jl = jm;
		else
			ju = jm;
		//cout << jl << " " << ju << endl;
		}
	//cout << jl << endl;
	return jl;
}

} // namespace DEP
