/**********************************************************************/
//
// FILE:    scanner.h
//          Scanner image class.
// CREATED: 24 May 02  (ds)
// CHANGES: 
// REFS:    
//
// (c) 2002 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#ifndef DEP_SCANNER_INCLUDED
#define DEP_SCANNER_INCLUDED

#include "frameext.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{
class CCDCamera;
class OrientationData;

/** Exterior orientation class.
 *  Assuming a perspective projection, the mapping from a 3D object coordinate 
 *  system onto the 2D image coordinate system is accomplished via the
 *  collinearity equation. The spatial position and attitude of the camera
 *  station is defined by the Exterior Orientation (EO). 
 *  The spatial position, also called perspective projection centre, is
 *  defined by its coordinates Xo, Yo and Zo. The attitude is defined by
 *  the three rotation angles. In this base class no assumtion is made on 
 *  the parametrisation of the three rotation angles. This spezialisation
 *  is done by the drived classes, e.g. ExtOrientationOPKRot.
 *
 *  @author Dirk Stallmann
 */
class ScannerImage : public ExtFrameImage
{
public:
	/// constructor
	ScannerImage();

	/// virtual destructor
	virtual ~ScannerImage() {}

	/**@name read/print image configuration
	 */
	//@{
		/// read image configuration
		bool read(istream& is, const vector<Camera*>& cams);

		/// print image configuration
		void print(ostream& os) const;
	//@}

	/// get image type
	string type_str() const;
	
	/**@name attach mandatory objects */
	//@{
		/// attach station
		bool attach(const vector<Station*>& stats)
		{
			assert(false);
			return false;
		}
	
		/// attach orientation data set
		bool attach(const vector<OrientationData*>& ods);
	//@}
	
	/// calculate exterior orientation
	bool calc_ext_orientation(DPoint3& pcc, DRotMatrix3& rm, double r) const;
	
	/// object-to-photo-coords. transformation
	bool object_to_photo(const DPoint3& obj_c, DPoint3& pho_c, double r) const;
	
	/**@name coordinate transformation */
	//@{
		/**@name object-to-image coords. transformation
		 *
		 *  @param obj_c (IN) object coordinates
		 *  @param img_c (OUT) image coordinates
		 *  @param r (IN) row/scanline coordinate
		 *  @return true = no problem, false = error encountered 
		 */
		bool object_to_image(const DPoint3& obj_c,
			DPoint2& img_c, double r) const;

		// image-to-object coords. transformation (ray-Z-plane intersection)
		//bool image_to_object(const DPoint2& img_c, DPoint3& obj_c) const;
	//@}
	
	bool observ_eqn_intersec(const string& id, DerivSub& a, DVector& wt,
		bool& found) const;
		
	/// add constribution to normal equations for all image point observations
	bool normal_eqns(const vector<ObjectPoint*>& obpts,
		Matrix<double>& atpa, Vector<double>& atpl, double& ltpl,
		int& num_obs) const;
		
		/// print residuals
// 	virtual void print_residuals(const vector<ObjectPoint*>& obj_pnts,
// 		DPoint2& sum_v, DPoint2& sum_vv, DPoint2& sum_pvv,
// 		DPoint2& sum_p, int& num, ostream& os) const;
		
	/// spatial intersection
	bool intersection(const DPoint2& ext_c1, const Image& image2,
		const DPoint2& ext_c2, DPoint3& obj_c, double& e) const;
	
protected:
	/// orientation data set ID
	string		_ods_id;
	
	/// orientation data (= scanner EO)
	OrientationData*	_ods;
	
	/** instantaneous exterior orientation parameters
	 *  position and attitude correction included
	 */
	//@{
		/// perspective centre coords.
		//DPoint3		_per_centre;

		/// rotation parameters
		//vector<double>	_rot_params;
		
		/** rotation matrix of R,
		 *  attitude correction included
		 */
		//DRotMatrix3	_R;
		
		/** rotation matrix of IMU misalignment M
		 */
		//DRotMatrix3	_M;
		
		/** rotation matrix: N = R M,
		 *  attitude correction and misalignment included
		 */
		//DRotMatrix3	_N;
	//@}
};


// non-member functions ...

/** ray-z-plane intersection
 *
 */
bool intersection(const CCDCamera& cam, const OrientationData& eo,
const DPoint2& pix_c, DPoint3& obj_c);


/** back projection by bisection method
 *
 */
bool projection_bisection(const CCDCamera& cam, const OrientationData& eo,
const DPoint3& obj_c, DPoint& pix_c,
int& nit, bool dbg=false);


/** back projection by hint method
 *
 */
bool projection_hint(const CCDCamera& cam, const OrientationData& eo,
const DPoint3& obj_c, DPoint& pix_c,
int& nit, double r0=-1, double eps=0.01, int max_iter=20,
bool dbg=false);

} // namespace DEP

#endif
