/**********************************************************************/
//
// FILE:    odf2txt.cpp
//          ODF (binary) to ASCII text conversion.
// CREATED: 10 Mar 04  (ds)
// CHANGES: 29 Mar 04  Read ODF versions 1.0 (= 0.0) and 1.1.(ds)
//          01 Jun 04  Calculation of some position statistics (ds).
//          08 Jun 04  Calculation of position and attitude range (ds).
// REFS:    ADS40 Info Kit
//
// (c) 2004 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#include <fstream>
#include "adsorient.h"

using namespace std;
using namespace DEP;


const char* const program_name = "odf2txt";
const char* const program_version = "version 0.5 (02/2007)";

static void usage(int status=0)
{
	cerr << "Usage: " << program_name << " [options] "
		"odf-file > txt-file\n"
		<< "Arguments:\n"
		<< "  odf-file         ODF filename.\n"
		<< "  txt-file         Text filename (optional).\n"
		<< "Options:\n"
		<< "  -hdr_only        Write only header information.\n"
		<< "  -range           Write range of position and attitude.\n"
		<< "  -stat            Write statistics of position changes.\n"
		<< "  -help            Show this text.\n"
		<< "  -verbose         Gives more information.\n"
		<< "  -version         Show program version.\n";

	exit(status);
}

static void error(const char* s, const char* s2 = "")
{
	cerr << program_name << ": " << s << ' ' << s2 << endl;
	exit(1);
}

// initialize static data members
double OrientationData::shift = 0.;		// global offset for orient. data
OrientationData::IpolMethod OrientationData::ipol = LINEAR;	// interpolation method

int main(int argc, char *argv[])
{
	if (argc == 1) usage(1);

	char*	odf_name = 0;		// input ODF filename
	//char*	txt_name = 0;		// output text filename
	bool	hdr_only = false;	// header only flag
	bool	range_flg = false;	// rnage flag
	bool	stat_flg = false;	// statitics flag
	bool	verb = false;		// verbose flag
	int		i = 0, j = 0;

	for (i = 1, j = 0; i < argc; i++)
		{
		if (**(argv + i) == '-')
			{
			if (strcmp(argv[i], "-help") == 0)
				usage(0);
			else if (strcmp(argv[i], "-hdr_only") == 0)
				hdr_only = true;
			else if (strcmp(argv[i], "-range") == 0)
				range_flg = true;
			else if (strcmp(argv[i], "-stat") == 0)
				stat_flg = true;
			else if (strcmp(argv[i], "-verbose") == 0)
				verb = true;
			else if (strcmp(argv[i], "-version") == 0)
				{
				cerr << program_name << " " << program_version
					<< " compiled " << __DATE__ << " " << __TIME__
					<< endl;
				exit(0);
				}
			else
				error("Illegal option:", argv[i]);
			}
		else
			{
			switch (j++)
				{
				case 0:
					odf_name = argv[i];
					break;

				//case 1:
				//	txt_name = argv[i];
				//	break;

				default:
					error("Too many arguments given:", argv[i]);
				}
			}
		}

	ADSOrientation eo;

	ifstream is(odf_name, ios_base::binary); // MS required binary open mode
	if (!is) error("Cannot open input ODF file:", odf_name);

	eo.read_header(is);

	eo.print_header(cout);
	cout << endl;
	if (hdr_only) return 0; // escape

	//ofstream os;
	//if (txt_name) os.open(odf_name);

	is.seekg(512, ios::beg);
	ADSOrientation::ODF odf;

	Form fix0(0); fix0.fixed().width(6);
	Form fix3(3); fix3.fixed().width(10);
	Form fix7(6); fix7.fixed().width(10);

	double xo_min = 0., yo_min = 0., zo_min = 0.;
	double xo_max = 0., yo_max = 0., zo_max = 0.;
	double a1_min = 0., a2_min = 0., a3_min = 0.;
	double a1_max = 0., a2_max = 0., a3_max = 0.;
	double xo_prev = 0., yo_prev = 0., zo_prev = 0.;
	double sum_dx = 0., sum_dy = 0., sum_dz = 0.;
	double dx_min = 0., dy_min = 0., dz_min = 0.;
	double dx_max = 0., dy_max = 0., dz_max = 0.;

	// read binary data and write as text
	// without unit conversion!!!
	cout << "row        time        Xo         Yo         Zo       omega       phi       kappa"
		<< "        s_Xo       s_Yo       s_Zo    s_omega    s_phi      s_kappa\n\n";
	for (i = 0; i < eo.get_numsets(); i++)
		{
		is.read((char*)&odf, sizeof(odf));
		//if (!is) error("Reading of orientation data failed.");

		double t = double(odf.t) / eo.get_tprec() + eo.get_basetime();
		double xo = double(odf.xo) / eo.get_pprec() + eo.get_basex();
		double yo = double(odf.yo) / eo.get_pprec() + eo.get_basey();
		double zo = double(odf.zo) / eo.get_pprec() + eo.get_basez();
		double a1 = double(odf.omega) / eo.get_aprec();
		double a2 = double(odf.phi) / eo.get_aprec();
		double a3 = double(odf.kappa) / eo.get_aprec();
		double sdev_xo = double(odf.sdev_xo) / eo.get_prec_sdev_pos();
		double sdev_yo = double(odf.sdev_yo) / eo.get_prec_sdev_pos();
		double sdev_zo = double(odf.sdev_zo) / eo.get_prec_sdev_pos();
		double sdev_a1 = double(odf.sdev_omega) / eo.get_prec_sdev_ang();
		double sdev_a2 = double(odf.sdev_phi) / eo.get_prec_sdev_ang();
		double sdev_a3 = double(odf.sdev_kappa) / eo.get_prec_sdev_ang();
		
		cout << fix0(i) << " " << fix3(t)
			<< " " << fix3(xo) << " " << fix3(yo) << " " << fix3(zo)
			<< " " << fix7(a1) << " " << fix7(a2) << " " << fix7(a3)
			<< " " << fix3(sdev_xo) << " " << fix3(sdev_yo) << " " << fix3(sdev_zo)
			<< " " << fix7(sdev_a1) << " " << fix7(sdev_a2) << " " << fix7(sdev_a3)
			<< endl;

		// range
		if (range_flg)
			{
			if (i == 0)
				{
				xo_min = xo_max = xo;
				yo_min = yo_max = yo;
				zo_min = zo_max = zo;
				a1_min = a1_max = a1;
				a2_min = a2_max = a2;
				a3_min = a3_max = a3;
				}
			else
				{
				if (xo < xo_min) xo_min = xo;
				if (yo < yo_min) yo_min = yo;
				if (zo < zo_min) zo_min = zo;
				if (xo > xo_max) xo_max = xo;
				if (yo > yo_max) yo_max = yo;
				if (zo > zo_max) zo_max = zo;
				if (a1 < a1_min) a1_min = a1;
				if (a2 < a2_min) a2_min = a2;
				if (a3 < a3_min) a3_min = a3;
				if (a1 > a1_max) a1_max = a1;
				if (a2 > a2_max) a2_max = a2;
				if (a3 > a3_max) a3_max = a3;
				}
			}

		// statistics
		if (stat_flg)
			{
			double dx = fabs(xo - xo_prev);
			double dy = fabs(yo - yo_prev);
			double dz = fabs(zo - zo_prev);

			if (i == 1)
				{
				dx_min = dx_max = dx;
				dy_min = dy_max = dy;
				dz_min = dz_max = dz;
				}
			else if (i > 1)
				{
				if (dx < dx_min) dx_min = dx;
				if (dy < dy_min) dy_min = dy;
				if (dz < dz_min) dz_min = dz;
				if (dx > dx_max) dx_max = dx;
				if (dy > dy_max) dy_max = dy;
				if (dz > dz_max) dz_max = dz;

				sum_dx += dx;
				sum_dy += dy;
				sum_dz += dz;
				}

			xo_prev = xo;
			yo_prev = yo;
			zo_prev = zo;
			}
		}

	// range
	if (range_flg)
		{
		cout << "\n\n";
		cout << "Range of position and attitude\n";
		cout << "------------------------------\n\n";
		cout << "              min        max      max-min\n\n";
		cout << "Xo     : " << fix3(xo_min) <<  " " << fix3(xo_max) << " " << fix3(xo_max - xo_min) << endl;
		cout << "Yo     : " << fix3(yo_min) <<  " " << fix3(yo_max) << " " << fix3(yo_max - yo_min) << endl;
		cout << "Zo     : " << fix3(zo_min) <<  " " << fix3(zo_max) << " " << fix3(zo_max - zo_min) << endl;
		cout << "omega  : " << fix7(a1_min) <<  " " << fix7(a1_max) << " " << fix7(a1_max - a1_min) << endl;
		cout << "phi    : " << fix7(a2_min) <<  " " << fix7(a2_max) << " " << fix7(a2_max - a2_min) << endl;
		cout << "kappa  : " << fix7(a3_min) <<  " " << fix7(a3_max) << " " << fix7(a3_max - a3_min) << endl;
		}

	// print statistics
	if (stat_flg)
		{
		int n = eo.get_numsets() - 1;
		double mean_dx = sum_dx / double(n);
		double mean_dy = sum_dy / double(n);
		double mean_dz = sum_dz / double(n);

		cout << "\n\n";
		cout << "Statistics of position changes\n";
		cout << "------------------------------\n";
		cout << "Absolute change of perspec. centre coords.\n\n";
		cout << "                DXo        DYo        DZo\n\n";
		cout << "Number   :  " << n << endl;
		cout << "Average  : " << fix3(mean_dx)
			<<  " " << fix3(mean_dy) << " " << fix3(mean_dz) << endl;
		cout << "Minimum  : " << fix3(dx_min)
			<<  " " << fix3(dy_min) << " " << fix3(dz_min) << endl;
		cout << "Maximum  : " << fix3(dx_max)
			<<  " " << fix3(dy_max) << " " << fix3(dz_max) << endl;
		cout << endl;
	}

	return 0;
}
