/* $Id: res2gplt.cpp,v 1.1.2.3 2008/03/05 18:28:54 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2008 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include "form.h"

using namespace std;
using namespace GenLib2;


const char* const program_name = "res2gplt";
const char* const program_version = "version 0.1.0 (03/2008)";

static void usage(int status=0)
{
	cerr << "Usage: " << program_name << " [options] "
		"res-file > gp-file\n"
		<< "Arguments:\n"
		<< "  res-file     Residuals of image coords. filename.\n"
		<< "  gp-file      Gnuplot filename.\n"
		<< "Options:\n"
		<< "  -help            Show this text.\n"
		<< "  -pointsize sz    Set point size. Default: 0.5.\n"
		<< "  -offset x y      Set label offset. Default: 0.5 0.5.\n"
		<< "  -vscale sc       Set vector scale factor. Default: 20.\n"
		<< "  -term <term>     Set terminal type = {png | eps}. Default: png.\n"
		<< "  -version         Show program version.\n";

	exit(status);
}

static void error(const char* s, const char* s2 = "")
{
	cerr << program_name << ": " << s << ' ' << s2 << endl;
	exit(1);
}


// image point residuals data record

struct ResData
{
	string	id;
	double	x, y;
	double	vx, vy;
};

// terminale
enum term_type
{
	TERM_PNG,
	TERM_EPS
};

int main(int argc, char *argv[])
{
	if (argc == 1) usage(1);
	
	term_type	term = TERM_PNG;
	char*	res_fname = 0;
	float	pointsize = 0.2;
	float	offset_x = 0.5;
	float	offset_y = 0.5;
	double	sc = 20.;
	int		i = 0, j = 0;
	
	for (i = 1, j = 0; i < argc; i++)
	{
		if (**(argv + i) == '-')
		{
			if (strcmp(argv[i], "-help") == 0)
				usage(0);
			else if (strcmp(argv[i], "-offset") == 0)
			{
				offset_x = atof(argv[++i]);
				offset_y = atof(argv[++i]);
			}
			else if (strcmp(argv[i], "-pointsize") == 0)
			{
				pointsize = atof(argv[++i]);
			}
			else if (strcmp(argv[i], "-vscale") == 0)
			{
				sc = atof(argv[++i]);
			}
			else if (strcmp(argv[i], "-term") == 0)
			{
				i++;
				if (strcmp(argv[i], "png") == 0)
					term = TERM_PNG;
				else if (strcmp(argv[i], "eps") == 0)
					term = TERM_EPS;
				else
					error("Illegal argument:", argv[i]);
			}
			else if (strcmp(argv[i], "-version") == 0)
			{
				cerr << program_name << " " << program_version
					<< " compiled " << __DATE__ << " " << __TIME__
					<< endl;
				exit(0);
			}
			else
				error("Illegal option:", argv[i]);
		}
		else
		{
			switch (j++)
			{
				case 0:
					res_fname = argv[i];
					break;

				default:
					error("Too many arguments given:", argv[i]);
			}
		}
	}
	
	vector<ResData*> res;

	ifstream is(res_fname);
	if (!is) error("Cannot open input image point residuals file:", res_fname);
	
	string out_fname(res_fname);
	size_t pos = out_fname.rfind('.');
	out_fname[pos] = '_';

	// Attention, this works only with cameraas with the same image size !!!!!
	string img_id;
	string id;
	double size_x = 0., size_y = 0.;
	ResData *rec = 0;
	while (is >> img_id)
	{
		if (img_id.compare("stop-dep") == 0) break;
		is >> size_x >> size_y;
		
		while (is >> id)
		{
			if (id.compare("stop-dep") == 0) break;
			
			rec = new(ResData);
			assert(rec != 0);

			rec->id = id;
			is >> rec->x >> rec->y;
			is >> rec->vx >> rec->vy;
			res.push_back(rec);
		}
	}
	
	vector<ResData*>::const_iterator p;
	
	switch (term)
	{
		case TERM_PNG:
			cout << "set term png size 900, 900" << endl;
			out_fname.append(".png");
			break;
					
		case TERM_EPS:
			cout << "set term postscript eps" << endl;
			out_fname.append(".eps");
			break;
		
		default:
			assert(false);
			break;
	}
	cout << "set output '" << out_fname << "'" << endl;
	//cout << "set autoscale xy" << endl;
// 	cout << "set autoscale x" << endl;
// 	cout << "unset autoscale y" << endl;
	//cout << "show pointsize" << endl;
	cout << "set pointsize " << pointsize << endl;
	cout << "set title 'Residuals of image coordinates - " << res_fname << "'" << endl;
	cout << "set xlabel 'x (um)'" << endl;
	cout << "set ylabel 'y (um)'" << endl;
	cout << "set format x '%.0f'" << endl;
	cout << "set format y '%.0f'" << endl;
	cout << "set key off" << endl;
	//cout << "set key title \"control\"" << endl;
	cout << "set grid" << endl;
	cout << "set datafile missing '?'" << endl;

	// line styles
	cout << "set style line 1 lt 1 lw 1" << endl;
	//cout << "set style line 2 lt 3 lw 2" << endl;

	// arrow styles
	cout << "set style arrow 1 nohead ls 1" << endl;
	//cout << "test" << endl;
	
	cout << "set size ratio -1" << endl;
	//cout << "set clip" << endl;
	cout << "set xr [" << (-size_x / 2.) << ":" << (size_x / 2.) << "]" << endl;
	cout << "set yr [" << (-size_y / 2.) << ":" << (size_y / 2.) << "]" << endl;

	Form fix9(3); fix9.fixed().width(9);
	
	cout << "plot '-' with points pt 5, \\" << endl;
	cout << "'-' with vectors arrowstyle 1" << endl;
	for (p = res.begin(); p != res.end(); ++p)
	{
		cout << fix9((*p)->x) << " " << fix9((*p)->y) << endl;
	}
	cout << "e" << endl;
	
	for (p = res.begin(); p != res.end(); ++p)
	{
		cout << fix9((*p)->x) << " " << fix9((*p)->y) << " ";
		cout << fix9((*p)->vx * sc) << " " << fix9((*p)->vy * sc) << endl;
	}
	cout << "e" << endl;
	
	return 0;
}
